"""
Common authorization token used for all authorized requests
"""


from dataclasses import dataclass
from datetime import datetime
from typing import Dict


@dataclass
class AuthorizationToken:
    type: str
    secret: str
    expires_at: datetime

    def is_expired(self) -> bool:
        return datetime.now() >= self.expires_at

    def to_authorization_header(self) -> Dict[str, str]:
        return {
            "Authorization": f"{self.type} {self.secret}",
        }

    def __str__(self) -> str:
        raise NotImplementedError



"""
Classes and methods to work with ViQube API
"""


from typing import Any, List, Optional, Set

import visiology_py as vi
import i2ls


class ApiV3(vi.BaseApi):
    def __init__(
        self,
        connection: vi.Connection,
    ) -> None:
        super().__init__(
            api_prefix="/viqube",
            api_version="3.0",
            authorization_scopes=[
                "openid",
                "profile",
                "email",
                "roles",
                "viqubeadmin_api",
                "viqube_api",
            ],
            authorization_headers={
                "Authorization": (
                    "Basic "
                    "dmlxdWJlYWRtaW5fcm9fY2xpZW50OjcmZEo1UldwVVMkLUVVQE1reHU="
                ),
                "Content-Type": "application/x-www-form-urlencoded",
            },
            connection=connection,
        )

    def version(
        self,
        token: Optional[vi.AuthorizationToken] = None,
    ) -> Any:
        return self._authorized_request(
            "GET",
            "/version",
            json=None,
            token=token,
        )

    def post_databases_query(
        self,
        database_unique_identifier: str,
        query: Any,
        token: Optional[vi.AuthorizationToken] = None,
    ) -> Any:
        return self._authorized_request(
            "POST",
            "/databases/{}/query".format(
                database_unique_identifier,
            ),
            json=query,
            token=token,
        )

    def get_databases_tables(
        self,
        database_unique_identifier: str,
        token: Optional[vi.AuthorizationToken] = None,
    ) -> Any:
        return self._authorized_request(
            "GET",
            "/databases/{}/tables".format(
                database_unique_identifier,
            ),
            json=None,
            token=token,
        )

    def get_databases_tables_records(
        self,
        database_unique_identifier: str,
        table_unique_identifier: str,
        token: Optional[vi.AuthorizationToken] = None,
    ) -> Any:
        return self._authorized_request(
            "GET",
            "/databases/{}/tables/{}/records".format(
                database_unique_identifier,
                table_unique_identifier,
            ),
            json=None,
            token=token,
        )

    def post_databases_tables_records(
        self,
        database_unique_identifier: str,
        table_unique_identifier: str,
        records: List[List[Any]],
        chunk_size: int = 100,
        token: Optional[vi.AuthorizationToken] = None,
    ) -> Any:
        # TODO: move chunk splitting to higher level api wrapper
        results = []
        for chunk in i2ls.chunks(records, chunk_size):
            results.append(self._authorized_request(
                "POST",
                "/databases/{}/tables/{}/records".format(
                    database_unique_identifier,
                    table_unique_identifier,
                ),
                json={
                    "values": chunk,
                },
                token=token,
            ))

        return results

    def delete_databases_tables_records(
        self,
        database_unique_identifier: str,
        table_unique_identifier: str,
        filter: Any = None,
        token: Optional[vi.AuthorizationToken] = None,
    ) -> Any:
        all_modifier = "/all" if filter is None else ""

        return self._authorized_request(
            "DELETE",
            "/databases/{}/tables/{}/records{}".format(
                database_unique_identifier,
                table_unique_identifier,
                all_modifier,
            ),
            json=filter,
            token=token,
        )

    def get_metadata_databases_dimensions(
        self,
        database_unique_identifier: str,
        token: Optional[vi.AuthorizationToken] = None,
    ) -> Any:
        return self._authorized_request(
            "GET",
            "/metadata/databases/{}/dimensions".format(
                database_unique_identifier
            ),
            json=None,
            token=token,
        )

    def get_metadata_databases_measuregroups(
        self,
        database_unique_identifier: str,
        token: Optional[vi.AuthorizationToken] = None,
    ) -> Any:
        return self._authorized_request(
            "GET",
            "/metadata/databases/{}/measuregroups".format(
                database_unique_identifier
            ),
            json=None,
            token=token,
        )

    def get_metadata_databases_measuregroups_dimensions(
        self,
        database_unique_identifier: str,
        measuregroup_unique_identifier: str,
        token: Optional[vi.AuthorizationToken] = None,
    ) -> Any:
        return self._authorized_request(
            "GET",
            "/metadata/databases/{}/measuregroups/{}/dimensions".format(
                database_unique_identifier,
                measuregroup_unique_identifier,
            ),
            json=None,
            token=token,
        )

    def get_metadata_databases_measuregroups_measures(
        self,
        database_unique_identifier: str,
        measuregroup_unique_identifier: str,
        token: Optional[vi.AuthorizationToken] = None,
    ) -> Any:
        return self._authorized_request(
            "GET",
            "/metadata/databases/{}/measuregroups/{}/measures".format(
                database_unique_identifier,
                measuregroup_unique_identifier,
            ),
            json=None,
            token=token,
        )

    def post_metadata_rawdata_query(
        self,
        query: Any,
        token: Optional[vi.AuthorizationToken] = None,
    ) -> Any:
        return self._authorized_request(
            "POST",
            "/metadata/rawdata/query",
            json=query,
            token=token,
        )


class Utils:
    def find_rawdata_column_id_by_name(
        rawdata: Any,
        column_name: str,
    ) -> Optional[int]:
        for id, column in enumerate(rawdata["columns"]):
            if column["header"] == column_name:
                return id

        return None

    def find_rawdata_distinct_values(
        rawdata: Any,
        column_name: str,
    ) -> Set[str]:
        result = set()

        column_id = Utils.find_rawdata_column_id_by_name(rawdata, column_name)

        for row in rawdata["values"]:
            result.add(row[column_id])

        return result



"""
Classes and methods to work with DataCollection API
"""


from typing import Any, Callable, Dict, List, Optional

import visiology_py as vi


class ApiV1(vi.BaseApi):
    def __init__(
        self,
        connection: vi.Connection,
    ) -> None:
        super().__init__(
            api_prefix="/datacollection/api",
            api_version="1.0",
            authorization_scopes=[
                "openid",
                "profile",
                "email",
                "roles",
                "viewer_api",
                "core_logic_facade",
            ],
            authorization_headers={
                "Authorization": "Basic cm8uY2xpZW50OmFtV25Cc3B9dipvfTYkSQ==",
                "Content-Type": "application/x-www-form-urlencoded",
            },
            connection=connection,
        )

    def get_dimension_attributes(
        self,
        dimension_id: int,
        token: Optional[vi.AuthorizationToken] = None,
    ) -> Any:
        return self._authorized_request(
            "GET",
            f"/dimensions/{dimension_id}/attributes?getAll=true",
            json=None,
            token=token,
        )

    def get_dimension_elements(
        self,
        dimension_id: int,
        filter: Any,
        token: Optional[vi.AuthorizationToken] = None,
    ) -> Any:
        return self._authorized_request(
            "GET",
            f"/dimensions/{dimension_id}/elements?getAll=true",
            json=filter,
            token=token,
        )

    def put_dimension_elements(
        self,
        dimension_id: int,
        elements: List[Any],
        token: Optional[vi.AuthorizationToken] = None,
    ) -> Any:
        return self._authorized_request(
            "PUT",
            f"/dimensions/{dimension_id}/elements",
            json=elements,
            token=token,
        )

    def post_dimension_elements(
        self,
        dimension_id: int,
        elements: List[Any],
        token: Optional[vi.AuthorizationToken] = None,
    ) -> Any:
        return self._authorized_request(
            "POST",
            f"/dimensions/{dimension_id}/elements",
            json=elements,
            token=token,
        )

    def delete_dimension_elements(
        self,
        dimension_id: int,
        filter: Any,
        token: Optional[vi.AuthorizationToken] = None,
    ) -> Any:
        return self._authorized_request(
            "DELETE",
            f"/dimensions/{dimension_id}/elements",
            json=filter,
            token=token,
        )


class ApiV2(vi.BaseApi):
    def __init__(
        self,
        connection: vi.Connection,
    ) -> None:
        super().__init__(
            api_prefix="/datacollection/api",
            api_version="2.0",
            authorization_scopes=[
                "openid",
                "profile",
                "email",
                "roles",
                "viewer_api",
                "core_logic_facade",
            ],
            authorization_headers={
                "Authorization": "Basic cm8uY2xpZW50OmFtV25Cc3B9dipvfTYkSQ==",
                "Content-Type": "application/x-www-form-urlencoded",
            },
            connection=connection,
        )

    def get_dimension_attributes(
        self,
        dimension_unique_name: str,
        token: Optional[vi.AuthorizationToken] = None,
    ) -> Any:
        return self._authorized_request(
            "GET",
            f"/dimensions/{dimension_unique_name}/attributes?getAll=true",
            json=None,
            token=token,
        )

    def get_dimension_elements(
        self,
        dimension_unique_name: str,
        filter: Any,
        token: Optional[vi.AuthorizationToken] = None,
    ) -> Any:
        return self._authorized_request(
            "GET",
            f"/dimensions/{dimension_unique_name}/elements?getAll=true",
            json=filter,
            token=token,
        )

    def put_dimension_elements(
        self,
        dimension_unique_name: str,
        elements: List[Any],
        token: Optional[vi.AuthorizationToken] = None,
    ) -> Any:
        return self._authorized_request(
            "PUT",
            f"/dimensions/{dimension_unique_name}/elements",
            json=elements,
            token=token,
        )

    def post_dimension_elements(
        self,
        dimension_unique_name: str,
        elements: List[Any],
        token: Optional[vi.AuthorizationToken] = None,
    ) -> Any:
        return self._authorized_request(
            "POST",
            f"/dimensions/{dimension_unique_name}/elements",
            json=elements,
            token=token,
        )

    def delete_dimension_elements(
        self,
        dimension_unique_name: str,
        filter: Any,
        token: Optional[vi.AuthorizationToken] = None,
    ) -> Any:
        return self._authorized_request(
            "DELETE",
            f"/dimensions/{dimension_unique_name}/elements",
            json=filter,
            token=token,
        )

    def get_measure_group_forms(
        self,
        measure_group_unique_name: str,
        filter: Any,
        token: Optional[vi.AuthorizationToken] = None,
    ) -> Any:
        return self._authorized_request(
            "GET",
            f"/measuregroups/{measure_group_unique_name}/forms",
            json=filter,
            token=token,
        )


class Utils:
    @staticmethod
    def find_attribute_by_unique_name(
        dimension_element: Any,
        attribute_unique_name: str,
    ) -> Any:
        for attribute in dimension_element["attributes"]:
            if attribute["attributeId"] == attribute_unique_name:
                return attribute

    @staticmethod
    def find_dimension_element_by_predicate(
        dimension_elements: Any,
        predicate: Callable[[Any], bool],
    ) -> Any:
        for element in dimension_elements["elements"]:
            if predicate(element):
                return element

    @staticmethod
    def find_dimension_element_by_attribute(
        dimension_elements: Any,
        attribute_unique_name: str,
        attribute_value: str
    ) -> Any:
        def predicate(element: Dict[str, Any]) -> bool:
            attribute = Utils.find_attribute_by_unique_name(
                element,
                attribute_unique_name,
            )
            return all((
                attribute is not None,
                attribute["value"] == attribute_value,
            ))

        return Utils.find_dimension_element_by_predicate(
            dimension_elements,
            predicate,
        )

    @staticmethod
    def find_dimension_element_by_name(
        dimension_elements: Any,
        element_name: str,
    ) -> Any:
        def predicate(element: Dict[str, Any]) -> bool:
            return bool(element["name"] == element_name)

        return Utils.find_dimension_element_by_predicate(
            dimension_elements,
            predicate,
        )

    @staticmethod
    def find_dimension_element_by_id(
        dimension_elements: Any,
        element_id: int,
    ) -> Any:
        def predicate(element: Dict[str, Any]) -> bool:
            return bool(element["id"] == element_id)

        return Utils.find_dimension_element_by_predicate(
            dimension_elements,
            predicate,
        )

    @staticmethod
    def prepare_dimension_element_to_insert(
        dimension_element: Any,
        attribute_map: Dict[str, Any],
    ) -> None:
        for attribute in dimension_element["attributes"]:
            attribute_id = attribute["attributeId"]
            if attribute_id in attribute_map:
                attribute["value"] = attribute_map[attribute_id]



"""
Base API class including methods shared between all APIs
"""


from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional

import requests

from visiology_py.authorization_token import AuthorizationToken
from visiology_py.connection import Connection


class BaseApi:
    def __init__(
        self,
        api_prefix: str,
        api_version: str,
        authorization_scopes: List[str],
        authorization_headers: Dict[str, str],
        connection: Connection,
    ) -> None:
        self._api_prefix = api_prefix
        self._api_version = api_version
        self._authorization_scopes = authorization_scopes
        self._authorization_headers = authorization_headers
        self._connection = connection

        self._token: Optional[AuthorizationToken] = None

    def _url(self, path: str) -> str:
        schema = self._connection.schema
        host = self._connection.host

        return f"{schema}://{host}{path}"

    def _prefixed_url(self, path: str) -> str:
        return self._url(f"{self._api_prefix}{path}")

    def _headers(self, token: AuthorizationToken) -> Dict[str, str]:
        return {
            **token.to_authorization_header(),
            "Content-Type": "application/json",
            "X-API-VERSION": self._api_version,
        }

    def emit_token(
        self,
    ) -> AuthorizationToken:
        response = requests.request(
            "POST",
            self._url("/idsrv/connect/token"),
            headers=self._authorization_headers,
            data={
                "grant_type": "password",
                "scope": " ".join(self._authorization_scopes),
                "response_type": "id_token token",
                "username": self._connection.username,
                "password": self._connection.password,
            },
        )

        token = response.json()
        expires_in = token["expires_in"]
        expires_at = datetime.now() + timedelta(seconds=expires_in)

        return AuthorizationToken(
            type=token["token_type"],
            secret=token["access_token"],
            expires_at=expires_at,
        )

    def _ensure_token(
        self,
    ) -> AuthorizationToken:
        if self._token is None or self._token.is_expired():
            self._token = self.emit_token()

        return self._token

    def _authorized_request(
        self,
        method: str,
        path: str,
        json: Any,
        token: Optional[AuthorizationToken] = None,
    ) -> Any:
        if token is None:
            token = self._ensure_token()

        response = requests.request(
            method,
            self._prefixed_url(path),
            headers=self._headers(token),
            json=json,
        )

        assert response.status_code == 200, response.text
        return response.json()



from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Hashable,
    Iterable,
    Iterator,
    List,
    Tuple,
    TypeVar,
)


__all__ = ["duplicates", "take", "chunks"]


T = TypeVar("T")


def duplicates(xs: Iterable[T]) -> Iterator[T]:
    """
    Returns generator of duplicates
    """

    seen: NohashSet[T] = NohashSet()
    yielded: NohashSet[T] = NohashSet()

    for x in xs:
        if x in seen and x not in yielded:
            yield x
            yielded.add(x)
        else:
            seen.add(x)


def take(i: Iterator[T], size: int) -> List[T]:
    """
    Gets iterator `i` and return a list that contains maximum `size` elements
    """

    result = []

    for _ in range(size):
        try:
            result.append(next(i))
        except StopIteration:
            pass

    return result


def chunks(xs: Iterable[T], size: int) -> Iterator[List[T]]:
    """
    Splits iterable into chunks as lists with lenght `size`.
    """

    iterator = iter(xs)

    chunk = take(iterator, size)
    while chunk:
        yield chunk
        chunk = take(iterator, size)


class NohashSet(Generic[T]):
    """
    Set-like collection with no `Hashable` restriction on elements
    """

    def __init__(self) -> None:
        self._real_buckets: Dict[int, List[T]] = dict()
        self._fallback_buckets: Dict[int, List[T]] = dict()

    def add(self, item: T) -> None:
        try:
            key = hash(item)
            if key not in self._real_buckets:
                self._real_buckets[key] = []

            self._real_buckets[key].append(item)
        except TypeError:
            pass

        key = hash(str(item))
        if key not in self._fallback_buckets:
            self._fallback_buckets[key] = []

        self._fallback_buckets[key].append(item)

    def __contains__(self, item: T) -> bool:
        if self._fast_real_buckets_check_not_contains(item):
            return False

        if self._fast_fallback_buckets_check_not_contains(item):
            return False

        if self._slow_real_buckets_check_contains(item):
            return True

        return self._slow_fallback_buckets_check_contains(item)

    def _fast_real_buckets_check_not_contains(self, item: T) -> bool:
        try:
            if hash(item) not in self._real_buckets:
                return True
        except TypeError:
            pass

        return False

    def _fast_fallback_buckets_check_not_contains(self, item: T) -> bool:
        return hash(str(item)) not in self._fallback_buckets

    def _slow_real_buckets_check_contains(self, item: T) -> bool:
        try:
            for item_in_bucket in self._real_buckets.get(hash(item), []):
                if item == item_in_bucket:
                    return True
        except TypeError:
            pass

        return False

    def _slow_fallback_buckets_check_contains(self, item: T) -> bool:
        for item_in_bucket in self._fallback_buckets.get(hash(str(item)), []):
            if item == item_in_bucket:
                return True

        return False



"""
Common Visiology's API classes and functions
"""


from visiology_py.authorization_token import AuthorizationToken
from visiology_py.base_api import BaseApi
from visiology_py.connection import Connection


__all__ = ["Connection", "AuthorizationToken", "BaseApi"]



import functools
import funcy
import inspect
import json
import time
from datetime import datetime, timedelta
from typing import Any, Callable, Dict, List, Type, TypeVar


Fany = Callable[..., Any]
Decorator = Callable[[Fany], Fany]


def timed(function: Fany) -> Fany:
    @functools.wraps(function)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        name = function.__name__
        start_time = time.time()
        result = function(*args, **kwargs)
        end_time = time.time()

        passed = (start_time - end_time) * 1000
        print(f"[{name}]: {passed:.2f} ms")

        return result

    return wrapper


# TODO: create a class instead of function, add idempotent_only mark
# TODO: add idempotent mark to api methods
# TODO: create and use NohashMap
# TODO: create and use Cache with TTL and LRU restrictions
def cached(time_to_live: timedelta) -> Decorator:
    whitelist = {
        # DC
        "get_dimension_attributes",
        "get_dimension_elements",

        # ViQube
        "version",
        "get_databases_tables",
        "get_databases_tables_records",
        "get_metadata_databases_dimensions",
        "get_metadata_databases_measuregroups",
        "get_metadata_databases_measuregroups_dimensions",
        "get_metadata_databases_measuregroups_measures",
        "post_databases_query",
        "post_metadata_rawdata_query",
    }

    def decorator(function: Fany) -> Fany:
        if function.__name__ not in whitelist:
            return function

        memo: Dict[str, Any] = {}

        @functools.wraps(function)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            key = json.dumps([args[1:], kwargs])
            now = datetime.now()

            if key not in memo:
                cached = False
            elif now > memo[key][1] + time_to_live:
                cached = False
            else:
                cached = True

            if cached:
                result, _ = memo[key]
            else:
                result = function(*args, **kwargs)
                memo[key] = result, now

            return result

        return wrapper

    return decorator


# TODO: allow to retry idempotent only or all methods
# TODO: check that max_tries more than zero
# TODO: allow max_tries to be None
def retried(
    max_tries: int,
    timeout_function: Callable[[int], float],
) -> Decorator:
    whitelist = {
        # DC
        "get_dimension_attributes",
        "get_dimension_elements",

        # ViQube
        "version",
        "emit_token",
        "get_databases_tables",
        "get_databases_tables_records",
        "get_metadata_databases_dimensions",
        "get_metadata_databases_measuregroups",
        "get_metadata_databases_measuregroups_dimensions",
        "get_metadata_databases_measuregroups_measures",
        "post_databases_query",
        "post_metadata_rawdata_query",
        "post_databases_tables_records",
        "delete_databases_tables_records",
    }

    def decorator(function: Fany) -> Fany:
        if function.__name__ not in whitelist:
            return function

        @functools.wraps(function)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            last_exception = Exception("Internal error in `retry` decorator")
            for try_number in range(max_tries):
                try:
                    return function(*args, **kwargs)
                except Exception as e:
                    time.sleep(timeout_function(try_number))
                    last_exception = e

            raise e

        return wrapper

    return decorator


def decorate_api(api: type, *decorators: Decorator) -> None:
    decorator = funcy.compose(*decorators)

    methods = inspect.getmembers(api, predicate=inspect.isfunction)
    for name, method in methods:
        setattr(api, name, decorator(method))



#!/usr/bin/env python3

from setuptools import find_packages, setup

setup(
    name="visiology-py",
    packages=["visiology_py", "i2ls"],
    version="0.4.0",
    description=(
        "High level wrappers for Visiology APIs: "
        "Datacollection, ViQube and ViQube admin"
    ),
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://gitlab.com/polymedia-orv/orv/visiology-py",
    author="Denis <codingjerk> Gruzdev",
    author_email="codingjerk@gmail.com",
    license="MIT",

    install_requires=[],
    setup_requires=[
        "dataclasses==0.6",
        "funcy==1.15",
        "hypothesis==5.36.0",
        "mypy==0.782",
        "pycodestyle==2.6.0",
        "pytest-cov==2.10.1",
        "pytest-runner==5.2",
        "pytest==6.0.1",
        "radon==4.3.2",
        "requests==2.24.0",
        "wheel==0.34.2",
    ],
    tests_require=[
        "hypothesis==5.36.0",
        "mypy==0.782",
        "pycodestyle==2.6.0",
        "pytest-cov==2.10.1",
        "pytest-runner==5.2",
        "pytest==6.0.1",
        "radon==4.3.2",
    ],
    test_suite="tests",
)
